from django.db import models
from django.utils import timezone
from django.contrib.auth import get_user_model

from phonenumber_field.modelfields import PhoneNumberField

from ..utils import random_code, valid_to


class SMSMessage(models.Model):
    """
    Save sended sms after as history
    """

    created = models.DateTimeField(auto_now_add=True)
    phone_number = models.CharField("Phone number", max_length=20)

    def __str__(self):
        return f"{self.phone_number} / {self.created}"

    def __repr__(self):
        return f"{self.phone_number}"

    class Meta:
        verbose_name = "Sms log"
        verbose_name_plural = "Sms log"


class PhoneCode(models.Model):
    """
    After validation save phone code instance
    """

    phone_number = PhoneNumberField(unique=True)
    owner = models.ForeignKey(get_user_model(),
                              null=True,
                              on_delete=models.CASCADE)
    code = models.PositiveIntegerField(default=random_code)
    valid_to = models.DateTimeField(default=valid_to)
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ("created_at",)
        verbose_name = "Phone code"
        verbose_name_plural = "Phone codes"

    def __str__(self):
        return f"{self.phone_number} ({self.code})"

    def __repr__(self):
        return self.__str__()

    @property
    def is_allow(self):
        return timezone.now() >= self.valid_to

    @property
    def message(self) -> str:
        return f"Your auth code: {self.code}"

    def save(self, *args, **kwargs):
        pretendent = self.__class__.objects.filter(
            phone_number=self.phone_number
        ).first()
        if pretendent is not None:
            self.pk = pretendent.pk

        super().save(*args, **kwargs)



from rest_framework import serializers

from ..conf import conf


def get_error(error_message):
    ERROR = {
        'message': '',
        'errors': [
            {
                "code": "0000",
                "field": None,
                "message": error_message
            }
        ]
    }

    return ERROR


class RestApiException(serializers.ValidationError):
    status_code = 400
    default_detail = ""


class SMSWaitException(RestApiException):
    default_detail = get_error(conf.SMS_WAIT_TIME)


class UserAlreadyExistException(RestApiException):
    default_detail = get_error(conf.SMS_USER_ALREADY_EXIST)


class SMSCodeNotFoundException(RestApiException):
    default_detail = get_error(conf.SMS_CODE_NOT_FOUND)



import importlib

from .conf import conf
from .models import PhoneCode


celery_conf = importlib.import_module(conf.SMS_CELERY_FILE_NAME)
app = getattr(celery_conf, "app")


def get_provider_class():
    provider = conf.SMS_PROVIDER

    return provider


@app.task
def send_sms_async(identifier: int):
    code_instance = PhoneCode.objects.filter(pk=identifier).first()
    if code_instance:
        provider_class = get_provider_class()
        provider = provider_class(
            to=code_instance.phone_number, message=code_instance.message, conf=conf
        )
        provider.send_sms()



from django.utils.module_loading import import_string

from rest_framework import generics, permissions

from ..conf import conf
from ..services import AuthService, GeneratorService
from .mixins import ResponsesMixin
from .serializers import \
    AuthSerializer, \
    EntrySerializer, \
    ChangePhoneNumberSerializer, \
    DefaultUserSerializer


class EntryAPIView(ResponsesMixin, generics.GenericAPIView):
    """
    Single endpoint to sign-in/sign-up
    :param
        - phone_number
    """

    permission_classes = [
        permissions.AllowAny,
    ]

    serializer_class = EntrySerializer

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        if serializer.is_valid():
            phone_number = serializer.validated_data.get("phone_number")
            GeneratorService.execute(phone_number=phone_number)
            return self.simple_text_response()
        else:
            return self.error_response(serializer.errors)


class AuthAPIView(ResponsesMixin, generics.GenericAPIView):
    """
    Single endpoint to auth thgrough phone_number + code
        params:
         - phone_number
         - code
    """

    permission_classes = [
        permissions.AllowAny,
    ]

    serializer_class = AuthSerializer

    def get_response_serializer(self):
        try:
            serializer = import_string(conf.SMS_USER_SERIALIZER)
        except ImportError:
            serializer = DefaultUserSerializer

        return serializer

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        if serializer.is_valid():
            phone_number = serializer.validated_data.get("phone_number")
            code = serializer.validated_data.get("code")
            user = AuthService.execute(phone_number=phone_number, code=code)
            serializer = self.get_response_serializer()
            success_value = serializer(instance=user, context={'request': request}).data

            return self.success_objects_response(success_value)
        else:
            return self.error_response(serializer.errors)


class ChangePhoneNumberAPIView(ResponsesMixin, generics.GenericAPIView):
    serializer_class = ChangePhoneNumberSerializer

    permission_classes = [
        permissions.IsAuthenticated,
    ]

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        if serializer.is_valid():
            new_phone_number = serializer.validated_data.get('new_phone_number')
            owner = request.user
            GeneratorService.execute(phone_number=new_phone_number, owner=owner)

            return self.simple_text_response()

        else:
            return self.error_response(serializer.errors)



from django.contrib.auth import get_user_model

from ..api.exceptions import \
    SMSWaitException, \
    UserAlreadyExistException

from ..models import PhoneCode
from ..utils import SmsService


class GeneratorService(SmsService):
    def __init__(self, phone_number: str, owner=None):
        self.phone_number = phone_number
        self.owner = owner

    def process(self):
        code = PhoneCode.objects\
            .filter(phone_number=self.phone_number)\
            .first()

        if code is not None:
            if not code.is_allow:
                raise SMSWaitException()

            code.delete()

        if self.owner is not None:
            if get_user_model().objects.filter(username=self.owner).exists():
                raise UserAlreadyExistException()

        PhoneCode.objects\
            .create(phone_number=self.phone_number,
                    owner=self.owner)
